/*
  This file is part of AyyiGtk. http://ayyi.org
  copyright (C) 2004-2012 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 3.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include <model/am_midi_part.h>
#include "am_part_manager.h"

static void am_note_selection_free(GList**);


MidiPart*
am_midi_part__new()
{
	MidiPart* part = g_new0(MidiPart, 1);

	AMPart* gpart = (AMPart*)part;

	gpart->ident.type = AYYI_OBJECT_MIDI_PART;

	pos_set(&gpart->length, 8, 0, 0);

	return part;
}


MidiNote*
am_midi_part__get_note_from_idx(AMPart* gpart, AyyiIdx shm_idx)
{
	AyyiMidiRegion* midi_region = ayyi_song__midi_region_at(gpart->ayyi->shm_idx);
	AyyiContainer* container = &midi_region->events;
	MidiNote* note = (MidiNote*)ayyi_song_container_get_item(container, shm_idx);
	return note;
}


MidiNote*
am_midi_part__get_next_event(AMPart* part, MidiNote* prev)
{
	g_return_val_if_fail(part, NULL);

	AyyiMidiRegion* region = ayyi_song__midi_region_at(part->ayyi->shm_idx);
	AyyiContainer* container = &region->events;
	if(container->last < 0) return NULL;

	if(prev && (am_midi_part__find_note(region, prev) >= AYYI_BLOCK_SIZE)) gwarn ("FIXME");

	int not_first = 0;

	int b; for(b=0;b<=container->last;b++){
		block* blk = ayyi_song__translate_address(container->block[b]);
		void** table = blk->slot;
		if(!table){ gerr ("shm error. table=null block=%i", b); return NULL; }

		if(!prev){
			//return first
			dbg (2, "first! region=%p", ayyi_song__translate_address(table[0]));
			return ayyi_song__translate_address(table[0]);
		}

		int i = (not_first++) ? 0 : am_midi_part__find_note(region, prev) + 1;
		for(;i<=blk->last;i++){
			dbg (2, "b=%i i=%i", b, i);
			MidiNote* next = ayyi_song__translate_address(table[i]);
			if(!next) continue;
			if(!is_song_shm_local(next)){ gerr ("bad shm pointer. table[%u]=%p", i, table[i]); return NULL; }
			return next;
		}
	}
	//if(!prev) dbg(0, "no events found.");
	return NULL;
}


MidiNote*
am_midi_part__get_prev_midi_event(AMPart* part, MidiNote* prev)
{
	AyyiMidiRegion* region = ayyi_song__midi_region_at(part->ayyi->shm_idx);
	AyyiContainer* container = &region->events;
	return ayyi_midi_note_prev(prev);
}


void
am_midi_part__print_events(AMPart* part)
{
	AyyiMidiRegion* region = ayyi_song__midi_region_at(part->ayyi->shm_idx);

	int n_events = ayyi_song_container_count_items(&region->events);
	if(n_events){
		printf("      events (%i):\n", n_events);
		int n = 0;
		MidiNote* note = NULL;
		while((note = (MidiNote*)ayyi_song_container_next_item(&region->events, note))){
			printf("        %i: %i %i %i %i\n", n, note->note, note->velocity, note->start, note->length);
			n++;
		}

		if(n != n_events) printf("        ??? total events: %i != %i\n", n, n_events);
	}
	else printf("      no events\n");
}


int
am_midi_part__find_note(AyyiMidiRegion* region, MidiNote* note)
{
	//temp? probably should add shm_idx to the note struct like the others.
	g_return_val_if_fail(region, false);
	g_return_val_if_fail(note, false);

	if(!&region->events){ gwarn("region has no events"); return -1; }

	return ayyi_container_find(&region->events, note, ayyi_song__translate_address);
}


gboolean
am_midi_part__sync(MidiPart* part)
{
	g_return_val_if_fail(part, false);
	gboolean err = false;
	AMPart* gpart = (AMPart*)part;
	g_return_val_if_fail(gpart->ayyi, false);

	AyyiMidiRegion* shared = ayyi_song__midi_region_at(gpart->ayyi->shm_idx);
	if(!shared->id){ gerr ("illegal id 0");/* err = true;*/ }

	_am_part_sync_name(gpart);
	_am_part_sync_pos(gpart);

	am_part__sync_length(gpart);

	/*
	if(shared->playlist_name && strlen(shared->playlist_name)){
		gpart->track = ardour_get_tracknum_from_playlist(shared->playlist_name);
		//gpart->track = part_get_tnum(gpart);
	} else {
		gerr ("part has no playlist! idx=%i", gpart->idx);
		err=true;
	}
	*/

	return !err;
}


MidiPart*
am_midi_part__init_from_index(MidiPart* mpart, AyyiIdx ayyi_index)
{
	/*
	initialise data for a part which is new to the gui but already exists in core.

	*/
	AMPart* part = (AMPart*)mpart;

	part->ayyi       = (AyyiRegionBase*)ayyi_song_container_get_item(&((AyyiSongService*)ayyi.service)->song->midi_regions, ayyi_index);
	part->ident.idx  = ayyi_index;
	part->ident.type = AYYI_OBJECT_MIDI_PART;
	part->track      = _am_part_get_ayyi_track(part);
	part->pool_item  = NULL;

	if(!part->track) return NULL;
	if(!am_midi_part__sync(mpart)) gerr ("init failed - should abort...\n");

	if(part->length.beat > MAX_PART_LENGTH_BEATS){
		gwarn ("part length too long (%i).\n", part->length.beat);
		pos_set(&part->length, 100, 0, 0);
	}

	am_parts__emit("add", part, NULL);

	return mpart;
}


void
am_midi_part__set_note_selection(MidiPart* part, GList* selection, gpointer user_data)
{
	//we take ownership of the selection.

	am_note_selection_free(&part->note_selection);
	part->note_selection = selection;

	//g_signal_emit_by_name (song, "note-selection-change", user_data);
	g_signal_emit_by_name (am_parts, "item-changed", part, CHANGE_SELECTION, user_data);

	dbg(0, "change=%i %i", CHANGE_SOLO, CHANGE_SELECTION);
}


void
am_midi_notes_modify_async(AMPart* gpart, const GList* notes)
{
	//@param notes, list of TransitionalNote*.

	void
	am_midi_notes_move_done(AyyiAction* action)
	{
		PF;
	}

	AyyiMidiRegion* region = ayyi_song__midi_region_at(gpart->ayyi->shm_idx); //errr... cant we just cast gpart->ayyi?

	AyyiAction* a  = ayyi_action_new("notes modify");
	a->callback    = am_midi_notes_move_done;

	GPtrArray* array = g_ptr_array_new();

	GList* l = (GList*)notes;
	int i = 0;
	for(;l;l=l->next){
		TransitionalNote* tn = l->data;
		MidiNote* note = &tn->transient;
		dbg(0, "idx=%i note=%i start=%u v=%i", note->idx, note->note, note->start, note->velocity);
		if(note->note > 128){ gwarn("bad note!"); continue; }
		if(note->length < 1){ gerr("bad note length!"); continue; }
		if(!ayyi_container_index_ok(&region->events, note->idx)){ gerr("bad note idx!"); continue; }

		GValue value = { 0, };
		g_value_init (&value, DBUS_STRUCT_UINT_BYTE_BYTE_UINT_UINT);
		g_value_take_boxed (&value, dbus_g_type_specialized_construct (DBUS_STRUCT_UINT_BYTE_BYTE_UINT_UINT));

		dbus_g_type_struct_set (&value, 0, note->idx, 1, note->note, 2, note->velocity, 3, note->start, 4, note->length, G_MAXUINT);
		g_ptr_array_add (array, g_value_get_boxed(&value));
		dbg(2, "array_len=%i key=%i", array->len, note->note);

		i++;
	}
	if(array->len) dbus_set_notelist(a, gpart->ayyi->shm_idx, array);

	g_ptr_array_free(array, TRUE);
	PF_DONE;
}


static void
am_note_selection_free(GList** selection)
{
	PF;
	if(*selection){
		GList* l = *selection;
		for(;l;l=l->next){
			//g_free(l->data); //TODO check
		}
#warning TODO make sure ownership is properly transferred
		//g_list_free(*selection);
		*selection = NULL;
	}
}


