/*
  This file is part of AyyiGtk. http://ayyi.org
  copyright (C) 2004-2012 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 3
  as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include <model/common.h>
#include <model/am_song.h>
#include "am_palette.h"

static void palette_from_style(Palette*, int i, gboolean bg, GtkStyle*, GtkStateType);


void
am_palette_load(Palette* palette)
{
  //initialise the colour palette.
  //the first 5 colours are the main gtk style colours (done separately)

  int i;
  for(i=4;i<MAX_COLOUR;i++){
    palette->red[i] = ((double)rand()) * 65535 / (RAND_MAX+1.0);
    palette->grn[i] = ((double)rand()) * 65535 / (RAND_MAX+1.0);
    palette->blu[i] = ((double)rand()) * 65535 / (RAND_MAX+1.0);
  }

  palette->red[5] = 0xffff;
  palette->grn[5] = 0xffff;
  palette->blu[5] = 0;
  palette->red[6] = 65535;
  palette->grn[6] = 0x6655;
  palette->blu[6] = 0;
  palette->red[7] = 65535;
  palette->grn[7] = 0;
  palette->blu[7] = 0;
  
  for(i=0;i<8;i++){
    palette->red[8+i] = i * 65535 / 8;
    palette->grn[8+i] = i * 65535 / 8;
    palette->blu[8+i] = 65535;
  }
  for(i=0;i<8;i++){
    palette->red[16+i] = i * 65535 / 8;
    palette->grn[16+i] = 65535;
    palette->blu[16+i] = i * 65535 / 8;
  }
  for(i=0;i<8;i++){
    palette->red[24+i] = 65535;
    palette->grn[24+i] = i * 65535 / 8;
    palette->blu[24+i] = i * 65535 / 8;
  }
  for(i=0;i<8;i++){
    palette->red[32+i] = 65535 -  i * 65535 / 8;
    palette->grn[32+i] = i * 65535 / 8;
    palette->blu[32+i] = 0;
  }
  for(i=0;i<8;i++){
    palette->red[40+i] = 0;
    palette->grn[40+i] = i * 65535 / 8;
    palette->blu[40+i] = 65535 -  i * 65535 / 8;
  }
  for(i=0;i<8;i++){
    palette->red[56+i] = i * 65535 / 8;
    palette->grn[56+i] = i * 65535 / 8;
    palette->blu[56+i] = i * 65535 / 8;
  }

  //TODO probably shouldnt emit this until the style colours are present
  // (they may or may not already be done depending on whether we are in app start-up)
  am_song__emit("palette-change");
}


void
palette_set_style_colours(Palette* palette, GtkStyle* style)
{
  //this has to be called by the application after the gtk widget styles have been initialised.

  gboolean bg = TRUE;
  gboolean fg = FALSE;
  palette_from_style(palette, 0, bg, style, GTK_STATE_SELECTED);
  palette_from_style(palette, 1, fg, style, GTK_STATE_NORMAL);
  palette_from_style(palette, 2, fg, style, GTK_STATE_SELECTED);
  palette_from_style(palette, 3, bg, style, GTK_STATE_NORMAL);

  GdkColor* colour = &style->base[GTK_STATE_SELECTED];
  palette->red[4] = colour->red;
  palette->grn[4] = colour->green;
  palette->blu[4] = colour->blue;

  am_song__emit("palette-change");
}


void
palette_lookup(Palette* palette, GdkColor* color, int c_idx)
{
  //gives us a gdkcolor from the given colour index number.
  
  if(c_idx > 63){
    gerr ("colouridx out of range: idx=%i.\n", c_idx);
    c_idx = 0;
  }
  color->red   = palette->red[c_idx];
  color->green = palette->grn[c_idx];
  color->blue  = palette->blu[c_idx];
}


#define ensure_colour_idx(c_idx) \
  if(c_idx < 0 || c_idx > 63){ \
    gerr ("colouridx out of range: %i", c_idx); \
    c_idx = 0; \
  }

uint32_t
palette_get_rgba(Palette* palette, int c_idx, unsigned char alpha)
{
  //gives us a color string from the given colour index number and alpha.

  ensure_colour_idx(c_idx);

  char red   = palette->red[c_idx]/256;
  char green = palette->grn[c_idx]/256;
  char blue  = palette->blu[c_idx]/256;

  uint32_t rgba = alpha + 256 * (blue + 256 * (green + 256 * red));

  return rgba;
}


void
palette_get_float(Palette* palette, int c_idx, float* r, float* g, float* b, const unsigned char alpha)
{
  if(c_idx > 63){ gerr ("colouridx out of range: %i\n", c_idx); c_idx = 0; }

  double _r = palette->red[c_idx];
  double _g = palette->grn[c_idx];
  double _b = palette->blu[c_idx];

  *r = _r / 0xffff;
  *g = _g / 0xffff;
  *b = _b / 0xffff;
}


void
am_rgba_to_float(uint32_t rgba, float* r, float* g, float* b)
{
	double _r = (rgba & 0xff000000) >> 24;
	double _g = (rgba & 0x00ff0000) >> 16;
	double _b = (rgba & 0x0000ff00) >>  8;

	*r = _r / 0xff;
	*g = _g / 0xff;
	*b = _b / 0xff;
	dbg (3, "%08x --> %.2f %.2f %.2f", rgba, *r, *g, *b);
}


uint32_t
am_colour_get_contrasting(Palette* palette, int colour_index)
{
  //return either black or white, to contrast best with the the given colour.

  //using b/w doesnt always work, as the contrast can be too high with at least some themes.

  //there is more than one approach we can take here.
  //-one is to use brighter darker values of the original colour.
  //-another is to find a suitable theme colour.

  int average = (palette->red[colour_index] + palette->grn[colour_index] + palette->blu[colour_index]) / 3;
  gboolean is_dark = (average < 0x7fff);
  uint32_t contrasting = average < 0x7fff ? 0xffffffff : 0x000000ff;
  dbg (3, "idx=%i av=%i", colour_index, average);

  if(palette->blu[colour_index]){//not black   //TODO make it work for other colours, and !is_dark
    if(is_dark){
      #define SATURATION 3 //unless we go into saturation, colours will not be bright enough.
      #define BLK_OFFSET1 20  //the colour furthest from the dominant colour needs to be boosted less.
      #define BLK_OFFSET2 40
      guint factor = ((0xffff * SATURATION) << 8 ) / palette->blu[colour_index]; //its possible that this may need to use the *average* colour

      contrasting = (MIN((((((palette->red[colour_index] >> 8) + BLK_OFFSET1) * factor) >> 8)), 0xff) << 24)
                  + (MIN((((((palette->grn[colour_index] >> 8) + BLK_OFFSET2) * factor) >> 8)), 0xff) << 16)
                  + (MIN((((((palette->blu[colour_index] >> 8)              ) * factor) >> 8)), 0xff) <<  8)
                  + 0xff;

      dbg(2, "factor=%i", factor);
    }
  }

  return contrasting;
}


static void
palette_from_style(Palette* palette, int i, gboolean bg, GtkStyle* style, GtkStateType state)
{
  GdkColor* colour;
  if(bg) colour = &style->bg[state];
  else   colour = &style->fg[state];

  palette->red[i] = colour->red;
  palette->grn[i] = colour->green;
  palette->blu[i] = colour->blue;
}


void
colour_get_style_base(GdkColor* color, int state)
{
  //get the default style base colour for the given widget state.

  GtkStyle* style = gtk_widget_get_default_style();
  color->red   = style->base[state].red;
  color->green = style->base[state].green;
  color->blue  = style->base[state].blue;
}


void
colour_get_style_bg_(GdkColor *color, int state)
{
  //get the default style background colour for the given widget state.

  GtkStyle* style = gtk_widget_get_default_style();
  color->red   = style->bg[state].red;
  color->green = style->bg[state].green;
  color->blue  = style->bg[state].blue;
}


void
colour_get_style_fg(GdkColor *color, int state)
{
  //get the default style foreground colour for the given widget state.

  GtkStyle* style = gtk_widget_get_default_style();
  color->red   = style->fg[state].red;
  color->green = style->fg[state].green;
  color->blue  = style->fg[state].blue;
}


void
colour_get_rgb_str(char* colorstr, int c_idx)
{
  //gives us a color string from the given colour index number.
  static GdkColor color;
  
  if(c_idx > 63){
    gerr ("colouridx out of range: %i", c_idx);
    c_idx = 0;
  }
  color.red   = song->palette.red[c_idx];
  color.green = song->palette.grn[c_idx];
  color.blue  = song->palette.blu[c_idx];

  snprintf(colorstr, 64, "#%02x%02x%02x", color.red/256, color.green/256, color.blue/256);
  dbg(3, "%i --> %s.", c_idx, colorstr);
}


void
colour_get_rgba_str(char* colorstr, int c_idx, const unsigned char alpha)
{
  //get a color string from the given colour index number and alpha.

  static GdkColor color;
  
  if(c_idx > 63){
    gerr ("colouridx out of range: %i", c_idx);
    c_idx = 0;
  }
  color.red   = song->palette.red[c_idx];
  color.green = song->palette.grn[c_idx];
  color.blue  = song->palette.blu[c_idx];

  snprintf(colorstr, 64, "0x%02x%02x%02x%02x", color.red/256, color.green/256, color.blue/256, alpha);
  dbg(0, "%s.\n", colorstr);
}


