/*
  This file is part of AyyiGtk.
  copyright (C) 2004-2013 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, version 3.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#define _GNU_SOURCE
#include "model/common.h"
#include <math.h>
#include <sndfile.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "waveform/waveform.h"
#include "waveform/wf_private.h" // for waveform_get_n_audio_blocks
#include <ayyi/ayyi_list.h>
#include <ayyi/ayyi_shm.h>
#include <ayyi/ayyi_client.h>
#include <ayyi/ayyi_song.h>
#include <ayyi/ayyi_log.h>
#include "model/am_pool.h"
#include "model/am_song.h"
#include "model/am_time.h"
#include "model/am_track.h"

#include "model/am_pool_item.h"

extern int    peak_mem_size;

static bool   pool_item_set_filename      (PoolItem*, const char* fname);
static gchar* pool_item_find_peakfile     (PoolItem*, int ch_num);
static bool   pool_item_generate_rms_file (PoolItem*, int ch_num);
static bool   pool_item_rms_file_is_valid (PoolItem*, const char* rms_file);
static void   pool_item_get_rms_filename  (char* peakname, const char* filename, int ch_num);
static void   pool_item_monitor_waveform  (PoolItem*);

//for rms calculatation:
enum sample_type {
    SAMPLE_TYPE_INT_8,
    SAMPLE_TYPE_INT_16,
    SAMPLE_TYPE_INT_32,
    SAMPLE_TYPE_FLOAT_32
};
typedef int AFframecount;
typedef int8_t rms_unit_t;
static void sample_calc_rms(rms_unit_t*, enum sample_type type, const void* buf, AFframecount count, float scale);


PoolItem*
pool_item_new()
{
	PoolItem* item       = g_new0(PoolItem, 1);
	item->state          = POOL_ITEM_NEW;
	item->valid          = true;
	item->peakfile_valid = false;
	item->pod_index[0]   = -1;
	item->pod_index[1]   = -1;
	item->max_db         = -100.0;

	return item;
}


void
pool_item_free(PoolItem* item)
{
	dbg(1, "item=%s", item->leafname);

	if(item->waveform){
		Waveform* waveform = item->waveform;
		item->waveform = NULL; // stop the finalise debugger complaining.
		g_object_unref(waveform);
	}else dbg(0, "%p ** no waveform to unref ! **", item);

	dbg (2, "freeing item...");
	g_free(item);
}


PoolItem*
pool_item_new_from_ayyi(AyyiFilesource* file)
{
	PoolItem* pool_item = pool_item_new();
	dbg(2, "item=%p", pool_item);
	pool_item->pod_index[0] = file->shm_idx;

	pool_item_sync(pool_item);

	if(pool_item_file_exists(pool_item)){
		if(!pool_item_set_file_info(pool_item)){
			pool_item_free(pool_item);
			return NULL;
		}
	}
	else gwarn ("file doesnt exist: dirname='%s' filename='%s'", pool_item->dirname, pool_item->filename);

	//check this is not a dupe:
	if(am_pool__item_exists(pool_item->filename)){
		dbg (2, "file is a dupe: '%s'", pool_item->filename); 
		pool_item_free(pool_item);
		return NULL;
	}

	char* fullpath = pool_item_get_full_path(pool_item, 0);
	pool_item->waveform = waveform_new(fullpath);
	pool_item_monitor_waveform(pool_item);

	if(!pool_item->state){
		pool_item_load_peak(pool_item, 0);
		pool_item->state = POOL_ITEM_OK;
	}

	g_free(fullpath);
	return pool_item;
}


static void
pool_item_monitor_waveform(PoolItem* pool_item)
{
	//temporarily used for debugging

	void pool_item_finalize_notify(gpointer _pi, GObject* was)
	{
		dbg(1, "waveform finalised! %s waveform=%p", ((PoolItem*)_pi)->leafname, ((PoolItem*)_pi)->waveform);
		//((PoolItem*)_pi)->waveform = NULL; // not needed
		g_return_if_fail(!((PoolItem*)_pi)->waveform);
	}

	g_object_weak_ref((GObject*)pool_item->waveform, pool_item_finalize_notify, pool_item);
}


void
pool_item_set_from_ayyi(PoolItem* item, AyyiFilesource* file)
{
	PF;
	pool_item_set_paths(item, file->name);

	pool_item_add_channel(item, file);

	AyyiFilesource* f2 = ayyi_file_get_linked(file);
	if(f2) pool_item_add_channel(item, f2);

	pool_item_sync(item);
	item->state = POOL_ITEM_OK;

	if(_debug_) pool_item_print(item);
}


gboolean
pool_item_set_file(PoolItem* pool_item, const char* fname, gboolean fast)
{
	pool_item_set_paths(pool_item, (char*)fname);

	if(!fast){
		if(g_file_test(fname, G_FILE_TEST_EXISTS)) pool_item_set_file_info(pool_item);
	}

	pool_item_set_filename(pool_item, fname);

	//check this is not a dupe:
	if(!fast){
		if(am_pool__item_exists(pool_item->filename)){
			dbg (0, "file is a dupe: %s. Aborting...", pool_item->filename); 
			return FALSE;
		}
	}

	if(!pool_item->waveform){
		pool_item->waveform = waveform_new(fname); //try and remove this
		pool_item_monitor_waveform(pool_item);
	}

	//initialise audio buffers for peakdata
	//TODO this should move to libwaveform?
#if 0
	if(!fast){
		waveform_get_n_audio_blocks(pool_item->waveform);
	}
#endif

	return TRUE;
}


static gboolean
pool_item_set_filename(PoolItem* pool_item, const char* fname)
{
	//pool_item->filename should be a relative path <--no that wont work. It should be either abs or relative.
	//                                                 -anyway, pool.h states that it should be the core name.

	//changelog: -we now update leafname and dirname here.
	//           -we now dont check the file exists. Should be done elsewhere.

	g_return_val_if_fail(strlen(fname), FALSE);
	g_return_val_if_fail(song->audiodir, FALSE);

	//if the file is within the audio root dir, store the relative path:
	if(fname[0]=='/'){
		dbg (2, "path is absolute.");
		if(strstr(fname, song->audiodir) == fname){
			gchar* relative = g_strdup_printf("%s", fname + strlen(song->audiodir) + 1); //+1 for the "/" char.
			dbg (0, "using relative path: %s.", relative);
			g_strlcpy(pool_item->filename, relative, 256);
			g_free(relative);
		}else{
			g_strlcpy(pool_item->filename, fname, 256);
		}

	}else{
		//path is not absolute.
		dbg (2, "path is not absolute.");
		g_strlcpy(pool_item->filename, fname, 256);
	}
	dbg (2, "filename=%s", pool_item->filename);

	pool_item_set_paths(pool_item, pool_item->filename);
	return true;
}


void
pool_item_add_channel(PoolItem* pool_item, AyyiFilesource* file)
{
	int c; for(c=0;c<2;c++){
		if(!pool_item->source_id[c]) break;
	}
	if(c > 1){ gerr ("channels full? ids: %Lu %Lu", pool_item->source_id[0], pool_item->source_id[1]); return; }

	pool_item->source_id[c] = file->id;
	pool_item->pod_index[c] = file->shm_idx;
	pool_item->channels     = c + 1;

	if(c) pool_item->waveform->is_split = true;

	pool_item_load_peak(pool_item, c);

	//pool_item_print(pool_item);
}


void
pool_item_set_paths(PoolItem* pool_item, char* path)
{
  g_return_if_fail(pool_item);
  dbg (2, "path=%s", path);

  if(path != pool_item->filename){ // dont pass pool_item->filename as an arg!
    strncpy(pool_item->filename, path, 255);
  }

  char* dirname = g_path_get_dirname(path);
  strncpy(pool_item->dirname, dirname, 255);
  g_free(dirname);

  audio_path_get_leaf(path, pool_item->leafname); 
}


gboolean
pool_item_set_file_info(PoolItem* pool_item)
{
	//uses libsndfile to get the number of samples, etc, in the given pool item.

	g_return_val_if_fail(pool_item, false);
	g_return_val_if_fail(pool_item->filename, false);

	if(pool_item_is_raw(pool_item)){
		gwarn ("FIXME file info not set for raw file.");
		return false;
	}

	char* filename = am_audio_path_get_full(pool_item->filename);

	SF_INFO sfinfo;
	sfinfo.format  = 0;

	SNDFILE* sffile;
	if(!(sffile = sf_open(filename, SFM_READ, &sfinfo))){
		pool_item->state = POOL_ITEM_INACCESSIBLE;
		gwarn ("%s (%s)", sf_strerror(NULL), filename);
		goto out;
	}

	if(sizeof(off_t) != 8) gerr("off_t=%i\n", sizeof(off_t));
	if(_debug_ > 1) printf("%iHz %s frames=%i\n", sfinfo.samplerate, am_format_channel_width(sfinfo.channels), (int)sfinfo.frames);
	if(!pool_item->channels) pool_item->channels = sfinfo.channels; // ->channels can be the combined width of more than one file.
	pool_item->sample_rate      = sfinfo.samplerate;
	pool_item->samplecount      = sfinfo.frames;
	pool_item->bytes_per_sample = 0; //FIXME

	if(pool_item->channels<1 || pool_item->channels>100){ gwarn ("bad channel count: %i", pool_item->channels); goto out; }
	if(sf_close(sffile)) gwarn ("bad file close.");

	g_free(filename);
	return true;

out:
	g_free(filename);
	return false;
}


void
pool_item_sync(PoolItem* item)
{
	g_return_if_fail(item->pod_index[0] > -1);
	AyyiFilesource* filesource = ayyi_song__filesource_at(item->pod_index[0]);
	g_return_if_fail(filesource);

	//dbg(0, "name=%s", filesource->name);
	char fname[256];
	g_strlcpy(fname, filesource->name, 256);
	//strip off the channel number from the end:
	gchar* pos = g_strrstr(fname, ":");
	if(pos) *pos = '\0';
	dbg (2, "%u: file='%s'", filesource->shm_idx, fname); 

	pool_item_set_file(item, fname, 1); //1 = fast

	item->source_id[0] = filesource->id;
	dbg (2, "setting source_id=%Lu (%s) size=%i", item->source_id[0], item->leafname, sizeof(*item));
}


gboolean
pool_item_file_exists(PoolItem* item)
{
  //does the file represented by this pool item exist in the file system?

  static char try_path[256];

  //try full path: FIXME item->filename may not be the full path.
  if(g_file_test(item->filename, G_FILE_TEST_EXISTS)) return true;

  //try audio_home directory:
  snprintf(try_path, 255, "%s/%s", song->audiodir, item->leafname);
  dbg (2, "try_path='%s'", try_path);
  if(g_file_test(try_path, G_FILE_TEST_EXISTS)) return true;

  return false;
}


static gchar*
pool_item_find_peakfile(PoolItem* pool_item, int ch_num)
{
	//return the peakfile name, and set the relevant pool_item properties.

	char leafname[256];
	if(ch_num){
		AyyiFilesource* filesource = ayyi_song__filesource_at(pool_item->pod_index[0]);
		char other[256];
		ayyi_file_get_other_channel(filesource, other, 256);
		strcpy(leafname, basename(other));
	}
	else strcpy(leafname, pool_item->leafname);

	char peakpath[256];
	snprintf(peakpath, 256, "%s/%s", song->peaksdir, leafname);

	//for some reason, peakfiles now have a "%A" in them!
	//(perhaps only if they are recorded rather than imported?)
	char* dot; if((dot = g_strrstr(peakpath, "."))) sprintf(dot, "%%A.");

	filename_change_extension(peakpath, "peak");
	if(!g_file_test(peakpath, G_FILE_TEST_EXISTS)){ 

		//try truncating at the first dot instead of the last.
		//-this seems wrong, but is apparently how Ardour works.
		snprintf(peakpath, 256, "%s/%s", song->peaksdir, leafname);
		dot = strstr(peakpath, ".");
		if(!dot) return NULL;
		*dot = '\0';
		filename_change_extension(peakpath, "peak");

		if(!g_file_test(peakpath, G_FILE_TEST_EXISTS)){ 
			//it is quite valid for there temporarily to be no peakfile.
			if(_debug_ > 1) gwarn ("cannot find peak file (%s)", peakpath);
			return NULL;
		}
	}

	pool_item->peakfile_valid  = true;

	return g_strdup(peakpath);
}


GList*
pool_item_get_regions(PoolItem* pool_item)
{
  /*
  returns a newly allocated list of idxs to songcore audio events (regions) for the given file.

  note: currently we get the regions directly from the engine - the local client does not independently store region info.

  */
  g_return_val_if_fail(pool_item, NULL);

  //find all the shared regions that have a source_id the same as our pool_item.
  GList* list = am_partmanager__get_regions_by_pool_item(pool_item);

  return list;
}


int
pool_item_count_regions(PoolItem* pool_item)
{
	g_return_val_if_fail(pool_item, 0);
	if(!pool_item->source_id){ gwarn ("pool item has missing id. Cannot get regions without id. pool_item='%s'", pool_item->leafname); return 0; }

	int count = 0;
	AyyiAudioRegion* region = NULL;
	while((region = ayyi_song__audio_region_next(region))){
		if(region->source0 == pool_item->source_id[0]){
			count++;
		}
	}
	dbg (2, "pool_item '%s' has %i regions.", pool_item->leafname, count);
	return count;
}


gboolean
pool_item_get_default_region(PoolItem* pool_item, uint32_t* region_idx)
{
	//return a reference to the region representing a whole file.

	//this assumes that the default region is first in the array. Strictly speaking i guess this isnt guaranteed.

	g_return_val_if_fail(pool_item, false);

#if 0
	struct block* block = ardour_region_get_block(0);
	int i;
	for(i=0;i<=block->last;i++){
		AyyiAudioRegion* region = ayyi_song__audio_region_at(i);
		if(!region) continue;

		if(region->source0 == pool_item->source_id[0]){
			*region_idx = i;
			return TRUE;
		}
	}
#endif
	AyyiAudioRegion* region = NULL;
	while((region = ayyi_song__audio_region_next(region))){
		if(region->source0 == pool_item->source_id[0]){
			*region_idx = region->shm_idx;
			return true;
		}
	}

	dbg (0, "no shared regions found for pool_item '%s'", pool_item->leafname);
	return 0;
}


gboolean
pool_item_is_raw(PoolItem* pool_item)
{
  char extn[8];
  //char* extn = pool_item->filename;
  filename_get_extension(pool_item->filename, extn);
  if(!strcmp(extn, "raw")) return TRUE;
  return FALSE;
}


int
pool_item_get_width(PoolItem* pool_item)
{
  if(pool_item->source_id[1]) return 2;
  if(pool_item->source_id[0]) return 1;
  return 0;
}


char*
pool_item_get_full_path(PoolItem* pool_item, int ch_num)
{
	//result must be free'd with g_free().

	char* src = NULL;
	char f[256];

	switch(ch_num){
		case 0:
			src = pool_item->filename;
			break;
		case 1:
			;AyyiFilesource* file = ayyi_song__filesource_at(pool_item->pod_index[0]);
			if(file){
				ayyi_file_get_other_channel(file, f, 256);
				src = f;
			}
			break;
		default:
			gwarn("bad channel number: %i", ch_num);
			return NULL;
	}

	char* fullpath = NULL;
	if(src[0]=='/'){
		//absolute path.
		fullpath = g_strdup(src);
	}else{
		//relative paths are relative to the song default audio directory.
		fullpath = g_strdup_printf("%s/%s", song->audiodir, src);
	}
	return fullpath;
}


void
pool_item_get_display_name(PoolItem* pool_item, char* name)
{
	char* d = replace(pool_item->leafname, "%L", "");
	strncpy(name, d, 255);
	d = replace(name, "-L.", ".");
	strncpy(name, d, 255);
}


gboolean
pool_item_in_core(PoolItem* item)
{
	//look through the Ayyi file list for the filesource_id of our pool_item.

	g_return_val_if_fail(item, false);

	AyyiFilesource* test = NULL;
	while((test = ayyi_song__filesource_next(test))){
		if(test->id == item->source_id[0]) return true;
	}

	return false;
}


bool
pool_item_is_nonempty(PoolItem* item, gpointer user_data)
{
	return item->samplecount;
}


bool
//pool_item__is_new(gpointer _item, gpointer user_data)
pool_item_is_new(PoolItem* item, gpointer user_data)
{
	//return pool items that are POOL_ITEM_NEW, or POOL_ITEM_PENDING

	//PoolItem* item = _item;
	return item->state <= POOL_ITEM_PENDING;
}


void
pool_item_load_peak(PoolItem* pool_item, int ch_num)
{
	//loads the peak file for the given poolitem into a buffer.

	char* filename = pool_item->filename;
	dbg (2, "ch=%i filename=%s", ch_num, filename);
	char* fullpath = pool_item_get_full_path(pool_item, 0);

	if(!g_file_test(fullpath, G_FILE_TEST_EXISTS)){ 
		gwarn ("audio file doesnt exist! (%s)", fullpath);
		pool_item->valid = false;
	} else pool_item->valid = true;

	pool_item->peakfile_valid = false; //the peak file can still be valid even if the pool_item is not.

	gchar* peak_file = pool_item_find_peakfile(pool_item, ch_num);
    if(!peak_file) goto out;

	pool_item->peakfile_valid = true;

	if(pool_item->waveform){
		waveform_load_peak(pool_item->waveform, peak_file, ch_num);

		pool_item->max_db = waveform_find_max_audio_level(pool_item->waveform);
		/*
		extern float int2db(short);
		dbg(0, "max_level=%.2fdB", int2db(pool_item->max_db));
		*/
	}

	//dbg (2, "done. %s: %isamples (%li beats / %.3f secs).", filename, pool_item->peak_buflen, samples2beats(pool_item->peak_buflen), samples2secs(pool_item->peak_buflen*WF_PEAK_RATIO));

	g_free(peak_file);

	pool_item_generate_rms_file(pool_item, ch_num);

  out:
	g_free(fullpath);
}


#if 0 //moved to libwaveform
RmsBuf*
pool_item_load_rms_file(PoolItem* pool_item, int ch_num)
{
	//loads the rms cache file for the given poolitem into a buffer.
	//Unlike peakfiles, these are not stored. g_free() the returned buffer after use.

	RmsBuf* rb = NULL;

#ifdef RMS_MMAP
  int fd;
  char* mmap_file;
  struct stat buf;
  //create the file
  if ((fd = open(argv[1], O_CREAT | O_RDWR, 0666)) < 0) {
    gerr("file open");
    return 3;
  }
  write(fd, starting_string, strlen(starting_string));
  //get size of file
  if (fstat(fd, &buf) < 0) {
    gerr("fstat error");
    return 4;
  }
  //create a buffer mapped to the file:
  if ((mmap_file = (char*) mmap(0, (size_t) buf.st_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0)) == (void*)-1) {
    gerr("mmap failure.");
    return;
  }
#endif

	dbg (2, "ch=%i filename=%s", ch_num, pool_item->filename);
	char* fullpath = pool_item_get_full_path((PoolItem*)pool_item, ch_num);

	//find the name of the peakfile:
	char rms_file[256];
	pool_item_get_rms_filename(rms_file, fullpath, ch_num);

	int fp = 0;
	if(!(fp = open(rms_file, O_RDONLY))){ gwarn ("file open failure."); goto out; }

	struct stat sinfo;
	if(stat(rms_file, &sinfo)){ gwarn ("rms file stat error. '%s'", rms_file); close(fp); goto out; }
	off_t fsize = sinfo.st_size;
	rb = g_new0(RmsBuf, 1);
	rb->size = fsize;
	rb->buf = g_new(char, fsize);

	//read the whole peak file into memory:
	if(read(fp, rb->buf, fsize) != fsize) gerr ("read error. couldnt read %Lu bytes from %s", fsize, rms_file);
  
	close(fp);
	//dbg (2, "done. %s: %isamples (%li beats / %.3f secs).", filename, pool_item->peak_buflen, samples2beats(pool_item->peak_buflen), samples2secs(pool_item->peak_buflen*WF_PEAK_RATIO));

	g_free(fullpath); //FIXME handle other returns
	return rb;

  out:
	g_free(fullpath);
	return rb;
}
#endif


const char*
pool_item_print_state(PoolItem* item)
{
	static char* states[] = {"POOL_ITEM_NEW", "POOL_ITEM_PENDING", "POOL_ITEM_BAD", "POOL_ITEM_INACCESSIBLE", "POOL_ITEM_OK"};
	g_return_val_if_fail(G_N_ELEMENTS(states) == POOL_ITEM_STATE_MAX, NULL);
	g_return_val_if_fail(item->state < POOL_ITEM_STATE_MAX, NULL);
	return states[item->state];
}


#if 0
const char*
pool_item_print_state(unsigned state)
{
  static char states[5][32] = {"POOL_ITEM_NEW", "POOL_ITEM_PENDING", "POOL_ITEM_BAD", "POOL_ITEM_INACCESSIBLE", "POOL_ITEM_OK" };
  if(state > 4) return NULL;
  return states[state];
}
#endif


void
pool_item_print(PoolItem* item)
{
	g_return_if_fail(item);
	char time[32]; format_time_samples(time, waveform_get_n_audio_blocks(item->waveform));
	printf("%s(): name=%s%s%s chs=%i idx=%i,%i buf=%s,%s state=%s samplecount=%Li nblocks=%i %s\n", __func__, bold, item->leafname, white, item->channels, item->pod_index[0], item->pod_index[1], waveform_peak_is_loaded(item->waveform, 0) ? "y" : "n", waveform_peak_is_loaded(item->waveform, 1) ? "y" : "n", pool_item_print_state(item), item->samplecount, waveform_get_n_audio_blocks(item->waveform), time);
}


#if 0
static void
peak_name(char* peakname, const char* filename, int ch_num)
{
	//currently we assume that the peakfile is in the same directory as the audio file,
	//and that it has the name extension .peak.

	//more versatile options will be added later.

	gchar* basename = g_path_get_basename(filename);

	gchar* path = g_build_filename(ayyi.service->song->path, ayyi.service->song->peaks_dir, basename, NULL);
	snprintf(peakname, 255, "%s%%A.peak", path);

	g_free(path);
	g_free(basename);
}
#endif


#define RMS_DATA_SIZE_BYTES 1
#define RMS_READ_SIZE 256

static bool
pool_item_generate_rms_file(PoolItem* pool_item, int ch_num)
{
	bool ok = true;
	char* src_path = pool_item_get_full_path(pool_item, ch_num);

	char rms_file[256];
	pool_item_get_rms_filename(rms_file, src_path, ch_num);
	dbg(2, "rms_file=%s", rms_file);

	if(pool_item_rms_file_is_valid(pool_item, rms_file)) goto out;

	SF_INFO       sfinfo;
	SNDFILE       *sffile;
	sfinfo.format = 0;

	int maxval = 0;

	if(!(sffile = sf_open(src_path, SFM_READ, &sfinfo))){
		log_print(LOG_FAIL, "not able to open file");
		dbg (0, "not able to open input file %s. %s", rms_file, sf_strerror(NULL));
		ok = false;
		goto out;
	}

	FILE* fp = fopen(rms_file, "wb");
	if(!fp){
		log_print(LOG_FAIL, "cannot open rms file for writing (%s).\n", rms_file);
		ok = false;
		goto out;
	}

	rms_unit_t out[RMS_READ_SIZE / WF_PEAK_RATIO];
    int readcount;
	int l = 0;
    int16_t* sf_data = g_malloc(sizeof(*sf_data) * RMS_READ_SIZE);
    float*  sf_dataf = g_malloc(sizeof(*sf_dataf) * RMS_READ_SIZE); //FIXME not freed?
	//while ((readcount = sf_read_short(sffile, sf_data, RMS_READ_SIZE))){
    while ((readcount = sf_read_float(sffile, sf_dataf, RMS_READ_SIZE))){
		//if(readcount < RMS_READ_SIZE) printf("EOF %i<%i\n", readcount, RMS_READ_SIZE);

		if(sf_error(sffile)){ gwarn("read error"); break; }

		int j; for(j=0;j<readcount;j++){
			sf_data[j] = sf_dataf[j] * (1 << 16);
		}

		//strictly speaking, the average should be calculated after display scaling - not here.
		float scale = WF_PEAK_RATIO;
		sample_calc_rms(out, SAMPLE_TYPE_INT_16, sf_data, RMS_READ_SIZE, scale);

		int write_size = MAX(readcount / WF_PEAK_RATIO, 1);
		if (fwrite(&out, RMS_DATA_SIZE_BYTES, write_size, fp) != write_size){
			gwarn("bad write.");
			break;
		}

		int i; for(i=0;i<readcount;i++){
			//maxval = MAX(out[i], maxval);
			maxval = MAX(sf_data[i] >> 8, maxval);
		}
		/*
		if (fwrite(&maxval, RMS_DATA_SIZE_BYTES, write_size, fp) != write_size){
			gwarn("bad write.");
			break;
		}
		*/
		dbg(3, "%i read=%i maxval=%i", l, readcount, maxval);
		l++;
	}
	//dbg(2, "maxlevel=%i,%i (%.3fdB)", totplus, totmin, int2db(MAX(totplus, -totmin)));
	g_free(sf_data);
	g_free(sf_dataf);

	if(sf_close(sffile)) gwarn ("bad file close.");
	fclose(fp);
  out:
	g_free(src_path);
	return ok;
}


//----------------------------------------------------------------


/* If not defined, sample_calc_rms calculates absolute averages. */

#define ENABLE_RMS

#ifndef ENABLE_RMS
#define RMS_INT(type, shift_c) \
    for(j = 0; j < q; j++) \
        sum += ABS(((type *)buf)[j]); \
    shift = shift_c

#else
#define RMS_INT(type, shift_c) \
    for(j = 0; j < q; j++) \
        fsum += (((type *)buf)[j] >> shift_c) * (((type *)buf)[j] >> shift_c)
#endif

#define HRES_MIN                      0.0625 /* inverse power of 2 */
#define HRES_MIN_SHIFT_BITS                4 /* bits to shift x to obtain ((1/x) * HRES_MIN)! */

static int
sample_get_width(enum sample_type type)
{
    switch(type) {
    case SAMPLE_TYPE_INT_8:
        return 1;
    case SAMPLE_TYPE_INT_16:
        return 2;
    case SAMPLE_TYPE_INT_32:
        return 4;
    case SAMPLE_TYPE_FLOAT_32:
        return sizeof(float);
    default:
        gerr("Unknown sample type!\n");
        return 0;
    }
}


/**
 * Calculates the root-mean-square for a range of samples.
 * @param rms Will receive the averages.
 * @param buf The samples.
 * @param count How many samples there are in buf.
 * @param type The type of the samples in buf.
 * @param scale The scaling factor: how many samples produce
 * a single average value.
 */
static void
sample_calc_rms(rms_unit_t* rms, enum sample_type type, const void* buf, AFframecount count, float scale)
{
#ifndef ENABLE_RMS
    long long sum = 0;
#endif
    double fsum = 0;
    int dst_idx = 0;
    int sample_width = sample_get_width(type);
    AFframecount i, j, q;
    AFframecount segment_size = scale * (int)(1 / HRES_MIN);
#ifndef ENABLE_RMS
    int shift = 0;
#endif

    count <<= HRES_MIN_SHIFT_BITS;

    for(i = 0; i < count; i += segment_size) {

        q = MIN(segment_size, count - i) >> HRES_MIN_SHIFT_BITS;
#ifndef ENABLE_RMS
        sum = 0;
#endif
        fsum = 0;

        switch(type) {
        case SAMPLE_TYPE_INT_8:
            RMS_INT(int8_t, 0);
            break;
        case SAMPLE_TYPE_INT_16:
            RMS_INT(int16_t, 8);
            break;
        case SAMPLE_TYPE_INT_32:
            RMS_INT(int32_t, 24);
            break;
        case SAMPLE_TYPE_FLOAT_32:
            for(j = 0; j < q; j++) 
#ifdef ENABLE_RMS
                fsum += ((float *)buf)[j] * ((float *)buf)[j];
#else
                fsum += ABS(((float *)buf)[j]);
#endif
            break;
        default:
            abort();
        }

#ifdef ENABLE_RMS
        if(type == SAMPLE_TYPE_FLOAT_32)
            rms[dst_idx++] = (int)(floor(sqrt(fsum / q) * 128));
        else
            rms[dst_idx++] = (int)(floor(sqrt(fsum / q)));
#else
        if(type == SAMPLE_TYPE_FLOAT_32)
            rms[dst_idx++] = (fsum / q) * 128;
        else 
            rms[dst_idx++] = (sum / q) >> shift;
#endif

        buf += (sample_width * q);
    }
    
}


static bool
pool_item_rms_file_is_valid(PoolItem* pool_item, const char* rms_file)
{
	// check the file exists and is dated newer than its source.
	// Outdated files are deleted.

	g_return_val_if_fail(rms_file, false);

	bool valid =true;

	if(!g_file_test(rms_file, G_FILE_TEST_EXISTS)){ 
		dbg (2, "rms_file doesnt exist. (%s)", rms_file);
		return false;
	}

	char* src_path = pool_item_get_full_path(pool_item, 0);
	if(am_file_is_newer(rms_file, src_path)) goto out;

	dbg(1, "rms_file is older - deleting... (%s)", rms_file);
	if(unlink(rms_file)) gwarn("failed to delete rms cache file.");
	valid = false;

  out:
	g_free(src_path);
	return valid;
}


static void
pool_item_get_rms_filename(char* peakname, const char* filename, int ch_num)
{
	dbg(2, "%i: %s", ch_num, filename);
	snprintf(peakname, 255, "%s.rms", filename);
	peakname[255] = '\0';
}


