/*
  copyright (C) 2012 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 3
  as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#ifndef __waveform_peak_h__
#define __waveform_peak_h__
#include <glib.h>
#include <glib-object.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include "stdint.h"
#include "waveform/typedefs.h"

G_BEGIN_DECLS

#define TYPE_WAVEFORM (waveform_get_type ())
#define WAVEFORM(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), TYPE_WAVEFORM, Waveform))
#define WAVEFORM_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), TYPE_WAVEFORM, WaveformClass))
#define IS_WAVEFORM(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), TYPE_WAVEFORM))
#define IS_WAVEFORM_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), TYPE_WAVEFORM))
#define WAVEFORM_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), TYPE_WAVEFORM, WaveformClass))

typedef struct _WaveformClass WaveformClass;

#define WF_PEAK_RATIO 256              // the number of samples per entry in the peakfile.
                                       // i.e. for every 4 bytes in peakfile, a 16bit audio file has 2*256 bytes
#define WF_PEAK_TEXTURE_SIZE 256       // the number of peakfile datapoints per texture.
                                       // i.e. 256 * 256 = 64k samples per texture, or 0.67 textures for 1 second of audio at 44.1k
#define WF_PEAK_VALUES_PER_SAMPLE 2    // one positive and one negative (unless using shaders).

#define WF_SHOW_RMS
#undef WF_SHOW_RMS

#define WF_PEAK_BLOCK_SIZE (256 * 256) // the number of audio frames per block (64k)

typedef int (*PeakLoader)(Waveform*, const char*, size_t);

enum
{
	WF_MONO = 1,
	WF_STEREO,
};

typedef struct
{
	uint64_t start;              // sample frames
	uint32_t len;                // sample frames
} WfSampleRegion;

typedef struct
{
	float left;
	float top;
	float len;
	float height;
} WfRectangle;

struct _Waveform
{
	GObject            parent_instance;

	char*              filename;          // either full path, or relative to cwd.
	uint64_t           n_frames;          // audio file size
	int                n_channels;
	gboolean           is_split;          // true for split stereo files

	gboolean           offline;

	GPtrArray*         hires_peaks;       // array of Peakbuf* TODO how much does audio_data deprecate this?
	int                num_peaks;         // peak_buflen / PEAK_VALUES_PER_SAMPLE
	RmsBuf*            rms_buf0;
	RmsBuf*            rms_buf1;

	WfGlBlock*         textures;          // opengl textures.
	WfGlBlock*         textures_lo;       // low res textures.
	WfTexturesHi*      textures_hi;

	//float            max_db;

	WaveformPriv*      priv;
};

struct _WaveformClass {
	GObjectClass parent_class;
};

typedef struct _buf_stereo_16
{
    short*     buf[WF_STEREO];
    guint      size;                      // number of shorts allocated, NOT bytes. When accessing, note that the last block will likely not be full.
    uint32_t   stamp;                     // put here for now. can move a parallel array if neccesary.
    uint64_t   start_frame;               // for debugging only
} WfBuf16;

struct _buf
{
	char* buf;
	guint size;
};

//high level api
Waveform*  waveform_load_new           (const char* filename);
void       waveform_set_peak_loader    (PeakLoader);
uint64_t   waveform_get_n_frames       (Waveform*);
int        waveform_get_n_channels     (Waveform*);

//low level api
GType      waveform_get_type           () G_GNUC_CONST;
Waveform*  waveform_new                (const char* filename);
Waveform*  waveform_construct          (GType);
#define    waveform_unref0(w)          (g_object_unref(w), w = NULL)
gboolean   waveform_load               (Waveform*);

gboolean   waveform_load_peak          (Waveform*, const char*, int ch_num);
gboolean   waveform_peak_is_loaded     (Waveform*, int ch_num);
RmsBuf*    waveform_load_rms_file      (Waveform*, int ch);

WfBuf16*   waveform_load_audio_async   (Waveform*, int block_num, int n_tiers_needed);
short      waveform_find_max_audio_level(Waveform*);

//TODO use #ifndef __GTK_H__
#ifdef USE_GDK_PIXBUF
#include <gtk/gtk.h>
//#else
//#ifndef GdkPixbuf
//#define GdkPixbuf void
//#endif
//#if !defined(GdkColor)
//#define GdkColor void
//#endif
#endif
typedef void (WfPixbufCallback)(Waveform*, GdkPixbuf*, gpointer);

void       waveform_peak_to_pixbuf       (Waveform*, GdkPixbuf*, WfSampleRegion*, uint32_t colour, uint32_t bg_colour);
void       waveform_peak_to_pixbuf_async (Waveform*, GdkPixbuf*, WfSampleRegion*, uint32_t colour, uint32_t bg_colour, WfPixbufCallback, gpointer);
void       waveform_peak_to_pixbuf_full  (Waveform*, GdkPixbuf*, uint32_t src_inset, int* start, int* end, double samples_per_px, uint32_t colour, uint32_t bg_colour, float gain);
void       waveform_rms_to_pixbuf        (Waveform*, GdkPixbuf*, uint32_t src_inset, int* start, int* end, double samples_per_px, GdkColor*, uint32_t bg_colour, float gain);

int32_t    wf_get_peakbuf_len_frames     ();

struct _colour_float
{
	float r;
	float g;
	float b;
};

#ifdef __wf_private__
#include "wf_private.h"
#endif

#endif //__waveform_peak_h__
